-- MINNIE.E - DOS32 svga mouse routines
-- Colin Taylor - cetaylor@compuserve.com
-- version 2.0 - 3/19/01

include machine.e
include graphics.e

integer video_mode,text_flag,visible,height,width,filter
sequence xor_map,and_map,hotspot,map,history,mouse_btn,last_btn,
    mouse_pos,last_pos
atom delay

function overlap(sequence pos1,sequence pos2)
    integer dx,dy,a,b,c,d
    dx = pos2[1]-pos1[1]
    dy = pos2[2]-pos1[2]
    if dx <= -width or dx >= width or dy <= -height or dy >= height then
	return 0
    end if
    if dx > 0 then
	a = 1+dx
	b = width
	c = 1
	d = width-dx
    else
	a = 1
	b = width+dx
	c = 1-dx
	d = width
    end if
    if dy > 0 then
	return {a,1+dy,b,height,c,1,d}
    else
	return {a,1,b,height+dy,c,1-dy,d}
    end if
end function

function get_image(sequence pos,integer w,integer h)
    sequence s
    pos &= w
    s = repeat(0,h)
    for i = 1 to h do
	s[i] = get_pixel(pos)
	pos[2] += 1
    end for
    return s
end function

procedure put_image(sequence pos,sequence s)
    for i = 1 to length(s) do
	pixel(s[i],pos)
	pos[2] += 1
    end for
end procedure

global procedure set_mouse_pos(sequence xy)
-- positions mouse at pixel coords {x,y}
    sequence regs
    regs = repeat(0,10)
    regs[REG_AX] = 4
    regs[REG_CX] = xy[1]
    regs[REG_DX] = xy[2]
    regs = dos_interrupt(#33,regs)
end procedure

global procedure limit_mouse(sequence min_xy,sequence max_xy)
-- sets min/max horizontal and vertical mouse position (pixel coords)
    sequence regs
    regs = repeat(0,10)
    regs[REG_AX] = 7
    regs[REG_CX] = min_xy[1]
    regs[REG_DX] = max_xy[1]
    regs = dos_interrupt(#33,regs)
    regs = repeat(0,10)
    regs[REG_AX] = 8
    regs[REG_CX] = min_xy[2]
    regs[REG_DX] = max_xy[2]
    regs = dos_interrupt(#33,regs)
end procedure

-- default cursor (arrow)
global sequence default_cursor
default_cursor = {{0,0,2,6,14,30,62,126,254,510,62,54,98,96,192,192,384,384,0},
    {2046,2044,2040,2032,2016,1984,1920,1792,1536,1024,0,1920,1800,1804,1566,
    1567,1087,1087,1663},{0,0}}
for i = 1 to 19 do
    default_cursor[1][i] = int_to_bits(default_cursor[1][i],11)*15
    default_cursor[2][i] = int_to_bits(default_cursor[2][i],11)*255
end for

procedure reset_mouse()
-- initializes minnie
    sequence regs,vc
    integer x,y
    vc = video_config()
    video_mode = vc[VC_MODE]
    regs = repeat(0,10)
    regs = dos_interrupt(#33,regs)
    x = vc[VC_XPIXELS]
    y = vc[VC_YPIXELS]
    visible = -1
    mouse_btn = {0,0,0}
    last_pos = {x,y}/2
    mouse_pos = last_pos
    filter = 127
    if find(video_mode,{0,1,2,3,7}) then
	text_flag = 1
    else
	text_flag = 0
	limit_mouse({0,0},{x,y}-1)  -- for svga
	set_mouse_pos({x,y}/2)  -- for svga
	xor_map = default_cursor[1]
	and_map = default_cursor[2]
	hotspot = default_cursor[3]
	height = length(xor_map)
	width = length(xor_map[1])
	map = and_map and not xor_map
	history = {}
	delay = 0.1
    end if
end procedure

reset_mouse()

procedure show_pointer(sequence pos)
    sequence back,temp
    integer a,b,c,d,e
    object o
    back = get_image(pos,width,height)
    temp = repeat(0,height)
    for i = 1 to height do
	temp[i] = xor_bits(and_bits(back[i],and_map[i]),xor_map[i])
    end for
    put_image(pos,temp)
    for i = 1 to length(history) do
	o = overlap(pos,history[i][2])
	if sequence(o) then
	    a = o[1]
	    b = o[3]
	    c = o[5]
	    d = o[7]
	    e = o[6]
	    for j = o[2] to o[4] do
		back[j][a..b] = history[i][3][e][c..d]
		e += 1
	    end for
	end if
    end for
    history &= {{time()+delay,pos,back}}
end procedure

global procedure mouse_pointer(integer show_it)
-- shows (1) or hides (0) the mouse pointer
    sequence vc,regs
    vc = video_config()
    if vc[VC_MODE] != video_mode then
	reset_mouse()
    end if
    if text_flag then
	regs = repeat(0,10)
	regs[REG_AX] = 2-(show_it > 0)
	regs = dos_interrupt(#33,regs)
    else
	if show_it then
	    if visible < 0 then
		visible += 1
	    end if
	    if visible = 0 then
		show_pointer(mouse_pos-hotspot)
	    end if
	else
	    if visible = 0 then
		for i = 1 to length(history) do
		    put_image(history[i][2],history[i][3])
		end for
		history = {}
	    end if
	    visible -= 1
	end if
    end if
end procedure

-- mouse button status
global constant LEFT_BTN = 1,RIGHT_BTN = 2,MIDDLE_BTN = 4

global function poll_mouse()
-- returns {button_status,x_coord,y_coord}
-- button_status = sum of (LEFT_BTN,RIGHT_BTN,MIDDLE_BTN)
    sequence regs,back,mask,nmask
    integer a,b,c,d,e
    object o
    last_btn = mouse_btn
    last_pos = mouse_pos
    regs = repeat(0,10)
    regs[REG_AX] = 3
    regs = dos_interrupt(#33,regs)
    mouse_btn = int_to_bits(regs[REG_BX],3)
    mouse_pos = {regs[REG_CX],regs[REG_DX]}
    if not text_flag then
	if visible = 0 then  -- mouse visible
	    if length(history) > 1 and history[1][1] <= time() then
		mask = map
		nmask = not mask
		for i = 2 to length(history) do
		    o = overlap(history[1][2],history[i][2])
		    if sequence(o) then
			a = o[1]
			b = o[3]
			c = o[5]
			d = o[7]
			e = o[6]
			for j = o[2] to o[4] do
			    mask[j][a..b] = mask[j][a..b]+nmask[e][c..d]
			    e += 1
			end for
		    end if
		end for
		back = get_image(history[1][2],width,height)
		for i = 1 to height do
		    for j = 1 to width do
			if mask[i][j] = 0 then
			    back[i][j] = history[1][3][i][j]
			end if
		    end for
		end for
		put_image(history[1][2],back)
		history = history[2..length(history)]
	    end if
	    if compare(mouse_pos,last_pos) then  -- mouse_moved
		show_pointer(mouse_pos-hotspot)
	    end if
	end if
    end if
    return regs[REG_BX] & mouse_pos
end function

-- mouse events
global constant MOVE = 1,LEFT_DOWN = 2,LEFT_UP = 4,RIGHT_DOWN = 8,
    RIGHT_UP = 16,MIDDLE_DOWN = 32,MIDDLE_UP = 64

global function get_mouse()
-- returns {event,x_coord,y_coord} or -1 if no event
    sequence void
    integer event
    void = poll_mouse()
    event = 0
    if compare(mouse_pos,last_pos) then
	event = 1  -- mouse moved
    end if
    if compare(mouse_btn,last_btn) then
	for i = 1 to 3 do  -- left,right,middle
	    if mouse_btn[i] and not last_btn[i] then
		event += power(2,i*2-1)  -- button pressed
	    elsif not mouse_btn[i] and last_btn[i] then
		event += power(2,i*2)  -- button released
	    end if
	end for
    end if
    event = and_bits(event,filter)
    if event then
	return event & mouse_pos
    else
	return -1
    end if
end function

global procedure mouse_events(integer events)
-- select the mouse events to be reported by get_mouse()
    if events < 0 then
	events = 0
    elsif events > 127 then
	events = 127
    end if
    filter = events
end procedure

global procedure set_mouse_cursor(sequence cursor)
-- graphic cursor = {xor_map,and_map,hotspot} or {} for default cursor
    if not text_flag then
	if length(cursor) = 0 then
	    cursor = default_cursor
	end if
	mouse_pointer(0)
	xor_map = cursor[1]
	and_map = cursor[2]
	hotspot = cursor[3]
	height = length(xor_map)
	width = length(xor_map[1])
	map = and_map and not xor_map
	history = {}
	mouse_pointer(1)
    end if
end procedure

